import paynt.verification.property
import payntbind
import stormpy

import pytest
from helpers.helper import get_sketch_paths, read_first_line
from math import inf


class TestSmgModelchecker:

    def test_simple_game_min(self):
        # setup
        smg_file, props_file = get_sketch_paths('smg/simple-game', props_name='circle-min.props')
        prop = read_first_line(props_file)
        program = stormpy.parse_prism_program(smg_file)
        properties = stormpy.parse_properties_for_prism_program(prop, program, None)
        model = stormpy.build_model(program, properties)
        paynt.verification.property.Property.initialize()

        # test
        result = payntbind.synthesis.model_check_smg(model, properties[0].raw_formula,
                                                    only_initial_states=False,
                                                    set_produce_schedulers=True,
                                                    env=paynt.verification.property.Property.environment)
        # assert
        assert result._quantitative
        assert result.result_for_all_states
        assert result.get_values() == pytest.approx([21, 20, 0])

        assert result.has_scheduler
        assert result.scheduler.deterministic
        assert result.scheduler.memoryless
        assert result.scheduler.get_choice(0).get_deterministic_choice() == 0
        assert result.scheduler.get_choice(1).get_deterministic_choice() == 1
        assert result.scheduler.get_choice(2).get_deterministic_choice() == 0

    def test_simple_game_max(self):
        # setup
        smg_file, props_file = get_sketch_paths('smg/simple-game', props_name='circle-max.props')
        prop = read_first_line(props_file)
        program = stormpy.parse_prism_program(smg_file)
        properties = stormpy.parse_properties_for_prism_program(prop, program, None)
        model = stormpy.build_model(program, properties)
        paynt.verification.property.Property.initialize()

        # test
        result = payntbind.synthesis.model_check_smg(model, properties[0].raw_formula,
                                                    only_initial_states=False,
                                                    set_produce_schedulers=True,
                                                    env=paynt.verification.property.Property.environment)
        # assert
        assert result._quantitative
        assert result.result_for_all_states
        assert result.get_values() == pytest.approx([12, 10, 0])

        assert result.has_scheduler
        assert result.scheduler.deterministic
        assert result.scheduler.memoryless
        assert result.scheduler.get_choice(0).get_deterministic_choice() == 1
        assert result.scheduler.get_choice(1).get_deterministic_choice() == 0
        assert result.scheduler.get_choice(2).get_deterministic_choice() == 0

    def test_simple_game_circle_ge(self):
        # setup
        smg_file, props_file = get_sketch_paths('smg/simple-game', props_name='circle-ge.props')
        prop = read_first_line(props_file)
        program = stormpy.parse_prism_program(smg_file)
        properties = stormpy.parse_properties_for_prism_program(prop, program, None)
        model = stormpy.build_model(program, properties)
        paynt.verification.property.Property.initialize()

        # test
        result = payntbind.synthesis.model_check_smg(model, properties[0].raw_formula,
                                                    only_initial_states=False,
                                                    set_produce_schedulers=True,
                                                    env=paynt.verification.property.Property.environment)
        # assert
        assert result._qualitative
        assert result.result_for_all_states
        assert not result.at(0)
        assert not result.at(1)
        assert not result.at(2)

    def test_simple_game_circle_le(self):
        # setup
        smg_file, props_file = get_sketch_paths('smg/simple-game', props_name='circle-le.props')
        prop = read_first_line(props_file)
        program = stormpy.parse_prism_program(smg_file)
        properties = stormpy.parse_properties_for_prism_program(prop, program, None)
        model = stormpy.build_model(program, properties)
        paynt.verification.property.Property.initialize()

        # test
        result = payntbind.synthesis.model_check_smg(model, properties[0].raw_formula,
                                                    only_initial_states=False,
                                                    set_produce_schedulers=True,
                                                    env=paynt.verification.property.Property.environment)
        # assert
        assert result._qualitative
        assert result.result_for_all_states
        assert not result.at(0)
        assert not result.at(1)
        assert result.at(2)

    def test_simple_game_square_ge(self):
        # setup
        smg_file, props_file = get_sketch_paths('smg/simple-game', props_name='square-ge.props')
        prop = read_first_line(props_file)
        program = stormpy.parse_prism_program(smg_file)
        properties = stormpy.parse_properties_for_prism_program(prop, program, None)
        model = stormpy.build_model(program, properties)
        paynt.verification.property.Property.initialize()

        # test
        result = payntbind.synthesis.model_check_smg(model, properties[0].raw_formula,
                                                    only_initial_states=False,
                                                    set_produce_schedulers=True,
                                                    env=paynt.verification.property.Property.environment)
        # assert
        assert result._qualitative
        assert result.result_for_all_states
        assert result.at(0)
        assert result.at(1)
        assert not result.at(2)

    def test_simple_game_square_le(self):
        # setup
        smg_file, props_file = get_sketch_paths('smg/simple-game', props_name='square-le.props')
        prop = read_first_line(props_file)
        program = stormpy.parse_prism_program(smg_file)
        properties = stormpy.parse_properties_for_prism_program(prop, program, None)
        model = stormpy.build_model(program, properties)
        paynt.verification.property.Property.initialize()

        # test
        result = payntbind.synthesis.model_check_smg(model, properties[0].raw_formula,
                                                    only_initial_states=False,
                                                    set_produce_schedulers=True,
                                                    env=paynt.verification.property.Property.environment)
        # assert
        assert result._qualitative
        assert result.result_for_all_states
        assert result.at(0)
        assert result.at(1)
        assert result.at(2)

    def test_distribution_min(self):
        # setup
        smg_file, props_file = get_sketch_paths('smg/distribution', props_name='circle-min.props')
        prop = read_first_line(props_file)
        program = stormpy.parse_prism_program(smg_file)
        properties = stormpy.parse_properties_for_prism_program(prop, program, None)
        model = stormpy.build_model(program, properties)
        paynt.verification.property.Property.initialize()

        # test
        result = payntbind.synthesis.model_check_smg(model, properties[0].raw_formula,
                                                    only_initial_states=False,
                                                    set_produce_schedulers=True,
                                                    env=paynt.verification.property.Property.environment)
        # assert
        assert result._quantitative
        assert result.result_for_all_states
        assert result.get_values() == pytest.approx([3, 2, 0])

        assert result.has_scheduler
        assert result.scheduler.deterministic
        assert result.scheduler.memoryless
        assert result.scheduler.get_choice(0).get_deterministic_choice() == 0
        assert result.scheduler.get_choice(1).get_deterministic_choice() == 0

    def test_distribution_max(self):
        # setup
        smg_file, props_file = get_sketch_paths('smg/distribution', props_name='circle-max.props')
        prop = read_first_line(props_file)
        program = stormpy.parse_prism_program(smg_file)
        properties = stormpy.parse_properties_for_prism_program(prop, program, None)
        model = stormpy.build_model(program, properties)
        paynt.verification.property.Property.initialize()

        # test
        result = payntbind.synthesis.model_check_smg(model, properties[0].raw_formula,
                                                    only_initial_states=False,
                                                    set_produce_schedulers=True,
                                                    env=paynt.verification.property.Property.environment)
        # assert
        assert result._quantitative
        assert result.result_for_all_states
        assert result.get_values() == pytest.approx([3, 2, 0])

        assert result.has_scheduler
        assert result.scheduler.deterministic
        assert result.scheduler.memoryless
        assert result.scheduler.get_choice(0).get_deterministic_choice() == 0
        assert result.scheduler.get_choice(1).get_deterministic_choice() == 0

    def test_medium_game_min(self):
        # setup
        smg_file, props_file = get_sketch_paths('smg/medium-game', props_name='circle-min.props')
        prop = read_first_line(props_file)
        program = stormpy.parse_prism_program(smg_file)
        properties = stormpy.parse_properties_for_prism_program(prop, program, None)
        model = stormpy.build_model(program, properties)
        paynt.verification.property.Property.initialize()

        # test
        result = payntbind.synthesis.model_check_smg(model, properties[0].raw_formula,
                                                    only_initial_states=False,
                                                    set_produce_schedulers=True,
                                                    env=paynt.verification.property.Property.environment)

        # assert
        assert result._quantitative
        assert result.result_for_all_states
        assert result.get_values() == pytest.approx([2.5, 1.5, 0, inf])

        assert result.has_scheduler
        assert result.scheduler.deterministic
        assert result.scheduler.memoryless
        assert result.scheduler.get_choice(0).get_deterministic_choice() == 0
        assert result.scheduler.get_choice(1).get_deterministic_choice() == 2
        assert result.scheduler.get_choice(2).get_deterministic_choice() == 0
        assert result.scheduler.get_choice(3).get_deterministic_choice() == 0

    def test_medium_game_max(self):
        # setup
        smg_file, props_file = get_sketch_paths('smg/medium-game', props_name='circle-max.props')
        prop = read_first_line(props_file)
        program = stormpy.parse_prism_program(smg_file)
        properties = stormpy.parse_properties_for_prism_program(prop, program, None)
        model = stormpy.build_model(program, properties)
        paynt.verification.property.Property.initialize()

        # test
        result = payntbind.synthesis.model_check_smg(model, properties[0].raw_formula,
                                                    only_initial_states=False,
                                                    set_produce_schedulers=True,
                                                    env=paynt.verification.property.Property.environment)

        # assert
        assert result._quantitative
        assert result.result_for_all_states
        assert result.get_values() == pytest.approx([inf, inf, 0, inf])

        assert result.has_scheduler
        assert result.scheduler.deterministic
        assert result.scheduler.memoryless
        assert result.scheduler.get_choice(0).get_deterministic_choice() == 0
        assert result.scheduler.get_choice(1).get_deterministic_choice() == 0
        assert result.scheduler.get_choice(2).get_deterministic_choice() == 0
        assert result.scheduler.get_choice(3).get_deterministic_choice() == 0

    def test_zero_reward_end_component_min(self):
        # setup
        smg_file, props_file = get_sketch_paths('smg/zero-reward-end-component', props_name='circle-min.props')
        prop = read_first_line(props_file)
        program = stormpy.parse_prism_program(smg_file)
        properties = stormpy.parse_properties_for_prism_program(prop, program, None)
        model = stormpy.build_model(program, properties)
        paynt.verification.property.Property.initialize()

        # test
        result = payntbind.synthesis.model_check_smg(model, properties[0].raw_formula,
                                                    only_initial_states=False,
                                                    set_produce_schedulers=True,
                                                    env=paynt.verification.property.Property.environment)

        # assert
        assert result.get_values() == pytest.approx([2, 1, 1, 0])

        assert result.scheduler.get_choice(0).get_deterministic_choice() == 0
        assert result.scheduler.get_choice(1).get_deterministic_choice() == 1
        assert result.scheduler.get_choice(2).get_deterministic_choice() == 0
        assert result.scheduler.get_choice(3).get_deterministic_choice() == 0

    def test_zero_reward_end_component_max(self):
        # setup
        smg_file, props_file = get_sketch_paths('smg/zero-reward-end-component', props_name='circle-max.props')
        prop = read_first_line(props_file)
        program = stormpy.parse_prism_program(smg_file)
        properties = stormpy.parse_properties_for_prism_program(prop, program, None)
        model = stormpy.build_model(program, properties)
        paynt.verification.property.Property.initialize()

        # test
        result = payntbind.synthesis.model_check_smg(model, properties[0].raw_formula,
                                                    only_initial_states=False,
                                                    set_produce_schedulers=True,
                                                    env=paynt.verification.property.Property.environment)

        # assert
        assert result.get_values() == pytest.approx([inf, inf, inf, 0])

        assert result.scheduler.get_choice(0).get_deterministic_choice() == 0
        assert result.scheduler.get_choice(1).get_deterministic_choice() == 0
        assert result.scheduler.get_choice(2).get_deterministic_choice() == 0
        assert result.scheduler.get_choice(3).get_deterministic_choice() == 0

    def test_self_loop_min(self):
        # setup
        smg_file, props_file = get_sketch_paths('smg/self-loop', props_name='circle-min.props')
        prop = read_first_line(props_file)
        program = stormpy.parse_prism_program(smg_file)
        properties = stormpy.parse_properties_for_prism_program(prop, program, None)
        model = stormpy.build_model(program, properties)
        paynt.verification.property.Property.initialize()

        # test
        result = payntbind.synthesis.model_check_smg(model, properties[0].raw_formula,
                                                    only_initial_states=False,
                                                    set_produce_schedulers=True,
                                                    env=paynt.verification.property.Property.environment)

        # assert
        assert result.get_values() == pytest.approx([inf, inf, 0])

        assert result.scheduler.get_choice(0).get_deterministic_choice() == 0
        assert result.scheduler.get_choice(1).get_deterministic_choice() == 0
        assert result.scheduler.get_choice(2).get_deterministic_choice() == 0

    def test_self_loop_max(self):
        # setup
        smg_file, props_file = get_sketch_paths('smg/self-loop', props_name='circle-max.props')
        prop = read_first_line(props_file)
        program = stormpy.parse_prism_program(smg_file)
        properties = stormpy.parse_properties_for_prism_program(prop, program, None)
        model = stormpy.build_model(program, properties)
        paynt.verification.property.Property.initialize()

        # test
        result = payntbind.synthesis.model_check_smg(model, properties[0].raw_formula,
                                                    only_initial_states=False,
                                                    set_produce_schedulers=True,
                                                    env=paynt.verification.property.Property.environment)

        # assert
        assert result.get_values() == pytest.approx([2, 1, 0])

        assert result.scheduler.get_choice(0).get_deterministic_choice() == 0
        assert result.scheduler.get_choice(1).get_deterministic_choice() == 1
        assert result.scheduler.get_choice(2).get_deterministic_choice() == 0

    def test_no_scheduler(self):
        # setup
        smg_file, props_file = get_sketch_paths('smg/simple-game', props_name='circle-min.props')
        prop = read_first_line(props_file)
        program = stormpy.parse_prism_program(smg_file)
        properties = stormpy.parse_properties_for_prism_program(prop, program, None)
        model = stormpy.build_model(program, properties)
        paynt.verification.property.Property.initialize()

        # test
        result = payntbind.synthesis.model_check_smg(model, properties[0].raw_formula,
                                                    only_initial_states=False,
                                                    set_produce_schedulers=False,
                                                    env=paynt.verification.property.Property.environment)

        # assert
        assert result.get_values() == pytest.approx([21, 20, 0])

        assert not result.has_scheduler


    def test_only_initial_states(self):
        # setup
        smg_file, props_file = get_sketch_paths('smg/simple-game', props_name='circle-min.props')
        prop = read_first_line(props_file)
        program = stormpy.parse_prism_program(smg_file)
        properties = stormpy.parse_properties_for_prism_program(prop, program, None)
        model = stormpy.build_model(program, properties)
        paynt.verification.property.Property.initialize()

        # test
        result = payntbind.synthesis.model_check_smg(model, properties[0].raw_formula,
                                                    only_initial_states=False,
                                                    set_produce_schedulers=True,
                                                    env=paynt.verification.property.Property.environment)
        # assert
        assert result._quantitative
        assert result.result_for_all_states
        assert result.get_values() == pytest.approx([21, 20, 0])

        assert result.has_scheduler
        assert result.scheduler.deterministic
        assert result.scheduler.memoryless
        assert result.scheduler.get_choice(0).get_deterministic_choice() == 0
        assert result.scheduler.get_choice(1).get_deterministic_choice() == 1
        assert result.scheduler.get_choice(2).get_deterministic_choice() == 0

    def test_only_initial_states(self):
        # setup
        smg_file, props_file = get_sketch_paths('smg/medium-game', props_name='circle-max.props')
        prop = read_first_line(props_file)
        program = stormpy.parse_prism_program(smg_file)
        properties = stormpy.parse_properties_for_prism_program(prop, program, None)
        model = stormpy.build_model(program, properties)
        paynt.verification.property.Property.initialize()

        # test
        result = payntbind.synthesis.model_check_smg(model, properties[0].raw_formula,
                                                    only_initial_states=True,
                                                    set_produce_schedulers=True,
                                                    env=paynt.verification.property.Property.environment)
        # assert
        assert result._quantitative
        assert result.result_for_all_states # other states will be filled anyway
        assert result.get_values() == pytest.approx([inf, inf, 0, inf])

        assert result.has_scheduler
        assert result.scheduler.deterministic
        assert result.scheduler.memoryless
        assert result.scheduler.get_choice(0).get_deterministic_choice() == 0
        assert result.scheduler.get_choice(1).get_deterministic_choice() == 0
        assert result.scheduler.get_choice(2).get_deterministic_choice() == 0
        assert result.scheduler.get_choice(3).get_deterministic_choice() == 0


    def test_stay_in_inf_state(self):
        # setup
        smg_file, props_file = get_sketch_paths('smg/stay-in-inf-state', props_name='circle-min.props')
        prop = read_first_line(props_file)
        program = stormpy.parse_prism_program(smg_file)
        properties = stormpy.parse_properties_for_prism_program(prop, program, None)
        model = stormpy.build_model(program, properties)
        paynt.verification.property.Property.initialize()

        # test
        result = payntbind.synthesis.model_check_smg(model, properties[0].raw_formula,
                                                    only_initial_states=False,
                                                    set_produce_schedulers=True,
                                                    env=paynt.verification.property.Property.environment)
        # assert
        assert result._quantitative
        assert result.result_for_all_states
        assert result.get_values() == pytest.approx([inf, inf, inf, 0])

        assert result.has_scheduler
        assert result.scheduler.deterministic
        assert result.scheduler.memoryless
        assert result.scheduler.get_choice(0).get_deterministic_choice() == 0
        assert result.scheduler.get_choice(1).get_deterministic_choice() == 0
        assert result.scheduler.get_choice(2).get_deterministic_choice() == 1
        assert result.scheduler.get_choice(3).get_deterministic_choice() == 0

    # model taken from prism examples
    def test_coins(self):
        # setup
        smg_file, props_file = get_sketch_paths('smg/coins')
        prop = read_first_line(props_file)
        program = stormpy.parse_prism_program(smg_file)
        properties = stormpy.parse_properties_for_prism_program(prop, program, None)
        model = stormpy.build_model(program, properties)
        paynt.verification.property.Property.initialize()

        # test
        result = payntbind.synthesis.model_check_smg(model, properties[0].raw_formula,
                                                    only_initial_states=False,
                                                    set_produce_schedulers=True,
                                                    env=paynt.verification.property.Property.environment)
        # assert
        assert result._quantitative
        assert result.result_for_all_states
        assert result.get_values() == pytest.approx([inf, 5, inf, 5, 5, inf, inf, 5, inf, inf, 5, 5, inf, inf, 5, 0, inf, inf, 0])
        # note: When obtaining state values from prism with -exportvector option, the values might be in a different order than expected.
        #    This is because the states are permuted in ConstructModel.java 497

        assert result.has_scheduler
        assert result.scheduler.deterministic
        assert result.scheduler.memoryless

    def test_probability(self):
        # setup
        smg_file, props_file = get_sketch_paths('smg/probabilities')
        prop = read_first_line(props_file)
        program = stormpy.parse_prism_program(smg_file)
        properties = stormpy.parse_properties_for_prism_program(prop, program, None)
        model = stormpy.build_model(program, properties)
        paynt.verification.property.Property.initialize()

        # test
        result = payntbind.synthesis.model_check_smg(model, properties[0].raw_formula,
                                                    only_initial_states=False,
                                                    set_produce_schedulers=True,
                                                    env=paynt.verification.property.Property.environment)
        # assert
        assert result._quantitative
        assert result.result_for_all_states
        assert result.get_values() == pytest.approx([0.5, 0.5, 1, 0])

        assert result.has_scheduler
        assert result.scheduler.deterministic
        assert result.scheduler.memoryless
        assert result.scheduler.get_choice(0).get_deterministic_choice() == 0
        assert result.scheduler.get_choice(1).get_deterministic_choice() == 1
        assert result.scheduler.get_choice(2).get_deterministic_choice() == 0
        assert result.scheduler.get_choice(3).get_deterministic_choice() == 0

    def test_probability_mec_exit(self):
        # setup
        smg_file, props_file = get_sketch_paths('smg/mec-exit')
        prop = read_first_line(props_file)
        program = stormpy.parse_prism_program(smg_file)
        properties = stormpy.parse_properties_for_prism_program(prop, program, None)
        model = stormpy.build_model(program, properties)
        paynt.verification.property.Property.initialize()

        # test
        result = payntbind.synthesis.model_check_smg(model, properties[0].raw_formula,
                                                    only_initial_states=False,
                                                    set_produce_schedulers=True,
                                                    env=paynt.verification.property.Property.environment)
        # assert
        assert result._quantitative
        assert result.result_for_all_states
        assert result.get_values() == pytest.approx([0.5, 0.5, 0, 1, 0.5, 0.5])

        assert result.has_scheduler
        assert result.scheduler.deterministic
        assert result.scheduler.memoryless
        assert result.scheduler.get_choice(0).get_deterministic_choice() == 2
        assert result.scheduler.get_choice(1).get_deterministic_choice() == 0
        assert result.scheduler.get_choice(2).get_deterministic_choice() == 0
        assert result.scheduler.get_choice(3).get_deterministic_choice() == 0
        assert result.scheduler.get_choice(4).get_deterministic_choice() == 1
        assert result.scheduler.get_choice(5).get_deterministic_choice() == 0


    def test_probability_circle_ge(self):
        # setup
        smg_file, props_file = get_sketch_paths('smg/probability-qualitative', props_name='circle-ge.props')
        prop = read_first_line(props_file)
        program = stormpy.parse_prism_program(smg_file)
        properties = stormpy.parse_properties_for_prism_program(prop, program, None)
        model = stormpy.build_model(program, properties)
        paynt.verification.property.Property.initialize()

        # test
        result = payntbind.synthesis.model_check_smg(model, properties[0].raw_formula,
                                                    only_initial_states=False,
                                                    set_produce_schedulers=True,
                                                    env=paynt.verification.property.Property.environment)
        # assert
        assert result._qualitative
        assert result.result_for_all_states
        assert not result.at(0)
        assert not result.at(1)
        assert result.at(2)

    def test_probability_circle_le(self):
        # setup
        smg_file, props_file = get_sketch_paths('smg/probability-qualitative', props_name='circle-le.props')
        prop = read_first_line(props_file)
        program = stormpy.parse_prism_program(smg_file)
        properties = stormpy.parse_properties_for_prism_program(prop, program, None)
        model = stormpy.build_model(program, properties)
        paynt.verification.property.Property.initialize()

        # test
        result = payntbind.synthesis.model_check_smg(model, properties[0].raw_formula,
                                                    only_initial_states=False,
                                                    set_produce_schedulers=True,
                                                    env=paynt.verification.property.Property.environment)
        # assert
        assert result._qualitative
        assert result.result_for_all_states
        assert not result.at(0)
        assert not result.at(1)
        assert not result.at(2)

    def test_probability_square_ge(self):
        # setup
        smg_file, props_file = get_sketch_paths('smg/probability-qualitative', props_name='square-ge.props')
        prop = read_first_line(props_file)
        program = stormpy.parse_prism_program(smg_file)
        properties = stormpy.parse_properties_for_prism_program(prop, program, None)
        model = stormpy.build_model(program, properties)
        paynt.verification.property.Property.initialize()

        # test
        result = payntbind.synthesis.model_check_smg(model, properties[0].raw_formula,
                                                    only_initial_states=False,
                                                    set_produce_schedulers=True,
                                                    env=paynt.verification.property.Property.environment)
        # assert
        assert result._qualitative
        assert result.result_for_all_states
        assert result.at(0)
        assert result.at(1)
        assert result.at(2)

    def test_probability_square_le(self):
        # setup
        smg_file, props_file = get_sketch_paths('smg/probability-qualitative', props_name='square-le.props')
        prop = read_first_line(props_file)
        program = stormpy.parse_prism_program(smg_file)
        properties = stormpy.parse_properties_for_prism_program(prop, program, None)
        model = stormpy.build_model(program, properties)
        paynt.verification.property.Property.initialize()

        # test
        result = payntbind.synthesis.model_check_smg(model, properties[0].raw_formula,
                                                    only_initial_states=False,
                                                    set_produce_schedulers=True,
                                                    env=paynt.verification.property.Property.environment)
        # assert
        assert result._qualitative
        assert result.result_for_all_states
        assert result.at(0)
        assert result.at(1)
        assert not result.at(2)


import stormpy
import payntbind

from helpers.helper import get_stormpy_example_path, get_builder_options

class TestMiscellaneous:

    def test_get_matrix_rows(self):

        builder_5x5 = stormpy.SparseMatrixBuilder(5, 5, force_dimensions=False)

        builder_5x5.add_next_value(0, 0, 0)
        builder_5x5.add_next_value(0, 1, 0.1)
        builder_5x5.add_next_value(2, 2, 22)
        builder_5x5.add_next_value(2, 3, 23)


        builder_5x5.add_next_value(3, 2, 32)
        builder_5x5.add_next_value(3, 4, 34)
        builder_5x5.add_next_value(4, 3, 43)

        matrix_5x5 = builder_5x5.build()

        rows = payntbind.synthesis.get_matrix_rows(matrix_5x5, [1, 2, 3])
        
        assert len(rows) == 3
        assert len(rows[0]) == 0

        entry_one = False
        for entry in rows[1]:
            if entry_one:
                assert entry.column == 3
                assert entry.value() == 23.0
            else	:
                assert entry.column == 2
                assert entry.value() == 22.0
                entry_one = True



import sys
import z3

# import pycvc5 if installed
import importlib
if importlib.util.find_spec('pycvc5') is not None:
    import pycvc5

import logging
logger = logging.getLogger(__name__)


class FamilyEncoding():

    def __init__(self, smt_solver, family):

        self.smt_solver = smt_solver
        self.family = family

        # for each hole, a formula encoding its possible options
        self.hole_clauses = None
        # SMT formula describing the family
        self.encoding = None
        # set to False as soon as pick_assignment returns None
        self.has_assignments = True

        hole_clauses = []
        for hole in range(family.num_holes):
            all_clauses = smt_solver.solver_clauses[hole]
            clauses = [all_clauses[option] for option in family.hole_options(hole)]
            if len(clauses) == 1:
                or_clause = clauses[0]
            else:
                if smt_solver.use_python_z3:
                    or_clause = z3.Or(clauses)
                elif smt_solver.use_cvc:
                    smt_solver = smt_solver.solver.mkTerm(pycvc5.Kind.Or, clauses)
                else:
                    pass
            hole_clauses.append(or_clause)

        if len(hole_clauses) == 1:
            encoding = hole_clauses[0]
        else:
            if smt_solver.use_python_z3:
                encoding = z3.And(hole_clauses)
            elif smt_solver.use_cvc:
                encoding = smt_solver.solver.mkTerm(pycvc5.Kind.And, hole_clauses)
            else:
                pass

        self.hole_clauses = hole_clauses
        self.encoding = encoding


    def pick_assignment(self):
        
        if not self.has_assignments:
            return None
        
        if self.smt_solver.use_python_z3:
            solver_result = self.smt_solver.solver.check(self.encoding)
            if solver_result == z3.unsat:
                self.has_assignments = False
                return None
            sat_model = self.smt_solver.solver.model()
            hole_options = []
            for hole_index,var in enumerate(self.smt_solver.solver_vars):
                option = sat_model[var].as_long()
                hole_options.append([option])
        elif self.smt_solver.use_cvc:
            solver_result = self.smt_solver.solver.checkSatAssuming(self.encoding)
            if solver_result.isUnsat():
                self.has_assignments = False
                return None
            hole_options = []
            for hole_index,var in enumerate(self.smt_solver.solver_vars):
                option = self.smt_solver.solver.getValue(var).getIntegerValue()
                hole_options.append([option])
        else:
            pass            
        
        assignment = self.family.assume_options_copy(hole_options)
        return assignment

        
class SmtSolver():

    def __init__(self, family):

        # SMT solver containing description of the unexplored design space
        self.solver = None
        # SMT solver choice
        self.use_python_z3 = False
        self.use_cvc = False
    
        # for each hole contains a corresponding solver variable
        self.solver_vars = None
        # for each hole contains a list of equalities [h==opt1,h==opt2,...],
        #   where h is the corresponding solver variable
        self.solver_clauses = None

        # current depth of push/pop solving
        self.solver_depth = 0

        # choose solver
        if "pycvc5" in sys.modules:
            logger.debug("using CVC5 for SMT solving.")
            self.use_cvc = True
        else:
            logger.debug("using Python Z3 for SMT solving.")
            self.use_python_z3 = True

        # create solver, solver variables
        self.solver_clauses = []
        if self.use_python_z3:
            self.solver = z3.Solver()
            self.solver_vars = [z3.Int(hole) for hole in range(family.num_holes)]
        elif self.use_cvc:
            self.solver = pycvc5.Solver()
            self.solver.setOption("produce-models", "true")
            self.solver.setOption("produce-assertions", "true")
            # self.solver.setLogic("ALL")
            # self.solver.setLogic("QF_ALL")
            self.solver.setLogic("QF_DT")
            # self.solver.setLogic("QF_UFDT")
            # self.solver.setLogic("QF_UFLIA")
            intSort = self.solver.getIntegerSort()
            self.solver_vars = [self.solver.mkConst(intSort, str(hole)) for hole in range(family.num_holes)]
        else:
            raise RuntimeError("Need to enable at least one SMT solver.")

        # create solver clauses
        self.solver_clauses = []
        for hole in range(family.num_holes):
            var = self.solver_vars[hole]
            clauses = [self.create_hole_clause(hole,option) for option in family.hole_options(hole)]
            self.solver_clauses.append(clauses)


    def create_hole_clause(self, hole, option):
        var = self.solver_vars[hole]
        if self.use_python_z3:
            return var == option
        elif self.use_cvc:
            return self.solver.mkTerm(pycvc5.Kind.Equal, var, self.solver.mkInteger(option))
        else:
            return None


    def pick_assignment(self, family):
        '''
        :return unexplored hole assignment from the family
            (or None if no instance remains)
        '''
        family.encode(self)
        return family.encoding.pick_assignment()

    def pick_assignment_priority(self, family, priority_subfamily):

        if priority_subfamily is None:
            return self.pick_assignment(family)

        # explore priority subfamily first
        assignment = self.pick_assignment(priority_subfamily)
        if assignment is not None:
            return assignment

        # explore remaining members
        return self.pick_assignment(family)
    
    
    def exclude_conflicts(self, family, assignment, conflicts):
        '''
        :param conflicts a list of conflicts (may be empty)
        :return estimate of pruned assignments
        '''
        pruning_estimate = 0
        for conflict in conflicts:
            pruning_estimate += self.exclude_conflict(family, assignment, conflict)
        return pruning_estimate
    
    
    def exclude_conflict(self, family, assignment, conflict):
        '''
        Exclude assignment from the family encoding using provided conflict.
        :param family base family
        :param assignment hole assignment that yielded unsatisfiable DTMC
        :param conflict indices of relevant holes in the corresponding counterexample
        :return estimate of pruned assignments
        '''
        assert family.encoding is not None
        
        if family.encoding is None:
            family.encoding = FamilyEncoding(self, family)

        pruning_estimate = 1
        counterexample_clauses = []
        for hole,var in enumerate(self.solver_vars):
            if hole in conflict:
                option = assignment.hole_options(hole)[0]
                counterexample_clauses.append(self.solver_clauses[hole][option])
            else:
                if family.hole_num_options(hole) < family.hole_num_options_total(hole):
                    counterexample_clauses.append(family.encoding.hole_clauses[hole])
                pruning_estimate *= family.hole_num_options(hole)

        if self.use_python_z3:
            if len(counterexample_clauses) == 0:
                counterexample_encoding = False
            else:
                counterexample_encoding = z3.Not(z3.And(counterexample_clauses))
            self.solver.add(counterexample_encoding)
        elif self.use_cvc:
            if len(counterexample_clauses) == 0:
                counterexample_encoding = self.solver.mkFalse()
            elif len(counterexample_clauses) == 1:
                counterexample_encoding = counterexample_clauses[0].notTerm()
            else:
                counterexample_encoding = self.solver.mkTerm(pycvc5.Kind.And, counterexample_clauses).notTerm()
            self.solver.assertFormula(counterexample_encoding)
        else:
            pass

        return pruning_estimate


    def level(self, refinement_depth):
        ''' Reset solver depth level to correspond to refinement level. '''

        if refinement_depth == 0:
            # fresh family, nothing to do
            return

        # reset to the scope of the parent (refinement_depth - 1)
        while self.solver_depth >= refinement_depth:
            self.solver.pop()
            self.solver_depth -= 1

        # create new scope
        self.solver.push()
        self.solver_depth += 1




import payntbind

import paynt.family.family
import paynt.quotient.quotient
import paynt.models.models

import json

import stormpy

import logging
logger = logging.getLogger(__name__)


class MdpFamilyQuotient(paynt.quotient.quotient.Quotient):

    # implicit size for scheduler memory unfolding
    initial_memory_size = 1

    @staticmethod
    def map_state_action_to_choices(mdp, num_actions, choice_to_action):
        state_action_choices = []
        for state in range(mdp.nr_states):
            action_choices = [[] for action in range(num_actions)]
            for choice in mdp.transition_matrix.get_rows_for_group(state):
                action = choice_to_action[choice]
                action_choices[action].append(choice)
            state_action_choices.append(action_choices)
        return state_action_choices

    @staticmethod
    def map_state_to_available_actions(state_action_choices):
        state_to_actions = []
        for state,action_choices in enumerate(state_action_choices):
            available_actions = []
            for action,choices in enumerate(action_choices):
                if choices:
                    available_actions.append(action)
            state_to_actions.append(available_actions)
        return state_to_actions

    
    def __init__(self, quotient_mdp, family, coloring, specification):

        self.memory_unfolder = None

        if MdpFamilyQuotient.initial_memory_size > 1:
            quotient_mdp, family, coloring = self.unfold_scheduler_memory(quotient_mdp, family, coloring)

        super().__init__(quotient_mdp = quotient_mdp, family = family, coloring = coloring, specification = specification)

        # number of distinct actions in the quotient
        self.num_actions = None
        # a list of action labels
        self.action_labels = None
        # for each choice of the quotient, the executed action
        self.choice_to_action = None
        # for each state of the quotient and for each action, a list of choices that execute this action
        self.state_action_choices = None
        # for each state of the quotient, a list of available actions
        self.state_to_actions = None

        self.action_labels,self.choice_to_action = payntbind.synthesis.extractActionLabels(quotient_mdp)
        self.num_actions = len(self.action_labels)
        self.state_action_choices = MdpFamilyQuotient.map_state_action_to_choices(
            self.quotient_mdp, self.num_actions, self.choice_to_action)
        self.state_to_actions = MdpFamilyQuotient.map_state_to_available_actions(self.state_action_choices)

    def empty_policy(self):
        return self.empty_scheduler()

    def scheduler_to_policy(self, scheduler, mdp):
        state_to_choice = self.scheduler_to_state_to_choice(mdp,scheduler)
        policy = self.empty_policy()
        for state in range(self.quotient_mdp.nr_states):
            choice = state_to_choice[state]
            if choice is not None:
                policy[state] = self.choice_to_action[choice]
        return policy

    def policy_to_state_valuation_actions(self, policy):
        '''
        Create a representation for a policy that associated action labels with state valuations. States with only
        one available action are omitted.
        '''
        policy,_ = policy
        sv = self.quotient_mdp.state_valuations
        state_valuation_to_action = []
        for state,action in enumerate(policy):
            if action is None:
                continue
            if len(self.state_to_actions[state])==1:
                continue
            # get action label
            action = self.action_labels[action]
            if action == "empty_label":
                continue

            # get state valuation
            valuation_jani = json.loads(str(sv.get_json(state)))
            valuation = {}
            for variable,value in valuation_jani.items():
                if "_loc_prism2jani_" in variable:
                    continue
                valuation[variable] = value

            state_valuation_to_action.append( (valuation,action) )

        # omit variables that are assigned to the same value
        default_valuation,_ = state_valuation_to_action[0]
        irrelevant_variables = set(default_valuation)
        for valuation,_ in state_valuation_to_action[1:]:
            for variable in list(irrelevant_variables):
                if valuation[variable] != default_valuation[variable]:
                    irrelevant_variables.remove(variable)
        state_valuation_to_action = [
            ({variable:value for variable,value in valuation.items() if variable not in irrelevant_variables},action)
            for valuation,action in state_valuation_to_action
        ]
        return state_valuation_to_action

    def policy_to_json(self, state_valuation_to_action, dt_control=False):
        '''
        :param state_valuation_to_action: a list of tuples (valuation,action) where valuation is a dictionary of variable
        :param dt_control: if True, outputs JSON in the format expected by the DT control tool,
                otherwise simpler format is used
        '''
        json_whole = []
        for index, valuation_action in enumerate(state_valuation_to_action):
            if dt_control:
                json_unit = {}
                valuation, action = valuation_action
                json_unit["c"] = [{"origin": {"action-label": action}}]
                json_unit["s"] = valuation
                json_whole.append(json_unit)
            else:
                json_whole.append(valuation_action)

        return json_whole

    
    def fix_and_apply_policy_to_family(self, family, policy):
        '''
        Apply policy to the quotient MDP for the given family. Every undefined action in a policy is set to an arbitrary
        one. Upon constructing the MDP, reset unused actions in a policy to None.
        :returns fixed policy
        :returns the resulting MDP
        '''
        policy = [action if action is not None else self.state_to_actions[state][0] for state,action in enumerate(policy)]
        policy_choices = []
        for state,action in enumerate(policy):
            policy_choices += self.state_action_choices[state][action]
        choices = payntbind.synthesis.policyToChoicesForFamily(policy_choices, family.selected_choices)

        # build MDP and keep only reachable states in policy
        mdp = self.build_from_choice_mask(choices)
        policy_fixed = self.empty_policy()
        for state in mdp.quotient_state_map:
            policy_fixed[state] = policy[state]

        mask = [state for state,action in enumerate(policy_fixed) if action is not None]
        policy_fixed = (policy_fixed,mask)
        return policy_fixed,mdp
    

    def apply_policy_to_family(self, family, policy):
        policy_choices = []
        for state,action in enumerate(policy):
            if action is None:
                for choice in self.state_action_choices[state]:
                    policy_choices += choice
            else:
                policy_choices += self.state_action_choices[state][action]
        choices = payntbind.synthesis.policyToChoicesForFamily(policy_choices, family.selected_choices)

        mdp = self.build_from_choice_mask(choices)

        return mdp

    
    def assert_mdp_is_deterministic(self, mdp, family):
        if mdp.is_deterministic:
            return
        
        logger.error(f"applied policy to a singleton family {family} and obtained MDP with nondeterminism")
        for state in range(mdp.model.nr_states):

            choices = mdp.model.transition_matrix.get_rows_for_group(state)
            if len(choices)>1:
                quotient_state = mdp.quotient_state_map[state]
                quotient_choices = [mdp.quotient_choice_map[choice] for choice in choices]
                state_str = self.quotient_mdp.state_valuations.get_string(quotient_state)
                state_str = state_str.replace(" ","")
                state_str = state_str.replace("\t","")
                actions_str = [self.action_labels[self.choice_to_action[choice]] for choice in quotient_choices]
                logger.error(f"the following state {state_str} has multiple actions {actions_str}")
        logger.error("aborting...")
        exit(1)
        

    def build_game_abstraction_solver(self, prop):
        target_label = prop.get_target_label()
        precision = paynt.verification.property.Property.model_checking_precision
        solver = payntbind.synthesis.GameAbstractionSolver(
            self.quotient_mdp, len(self.action_labels), self.choice_to_action, prop.formula, prop.maximizing, target_label, precision
        )
        return solver

    def build_assignment(self, family):
        assert family.size == 1, "expecting family of size 1"
        choices = self.coloring.selectCompatibleChoices(family.family)
        model,state_map,choice_map = self.restrict_quotient(choices)
        return paynt.models.models.SubMdp(model,state_map,choice_map)
    
    def unfold_scheduler_memory(self, quotient_mdp, family, coloring):
        '''
        Unfold the scheduler memory of the quotient MDP to the initial_memory_size.
        :returns a new quotient MDP with unfolded scheduler memory
        '''

        logger.info(f"unfolding scheduler memory of {self.initial_memory_size} into the model.")

        # unfold the scheduler memory into the model
        self.memory_unfolder = payntbind.synthesis.MemoryUnfolder(
            quotient_mdp
        )
        unfolded_mdp = self.memory_unfolder.construct_unfolded_model(MdpFamilyQuotient.initial_memory_size)

        # create new coloring
        choice_to_hole_options = []
        original_choice_to_hole_options = coloring.getChoiceToAssignment()
        choice_map = list(self.memory_unfolder.choice_map)
        for choice in range(unfolded_mdp.nr_choices):
            original_choice = choice_map[choice]
            choice_to_hole_options.append(original_choice_to_hole_options[original_choice])
     
        new_coloring = payntbind.synthesis.Coloring(family.family, unfolded_mdp.nondeterministic_choice_indices, choice_to_hole_options)

        logger.info(f"unfolded model has {unfolded_mdp.nr_states} states and {unfolded_mdp.nr_choices} choices.")

        return unfolded_mdp, family, new_coloring



import payntbind

import paynt.synthesizer.conflict_generator.dtmc
import paynt.verification.property

import logging
logger = logging.getLogger(__name__)


class ConflictGeneratorMdp(paynt.synthesizer.conflict_generator.dtmc.ConflictGeneratorDtmc):

    def initialize(self):
        state_to_holes_bv = self.quotient.coloring.getStateToHoles().copy()
        state_to_holes = []
        for state,holes_bv in enumerate(state_to_holes_bv):
            holes = set([hole for hole in holes_bv])
            state_to_holes.append(holes)
        formulae = self.quotient.specification.stormpy_formulae()
        self.counterexample_generator = payntbind.synthesis.CounterexampleGeneratorMdp(
            self.quotient.quotient_mdp, self.quotient.family.num_holes,
            state_to_holes, formulae
        )

    def prepare_model(self, model):
        self.counterexample_generator.prepare_mdp(model.model, model.quotient_state_map)



import stormpy
import payntbind
import math
import operator

import logging
logger = logging.getLogger(__name__)


def construct_property(prop, relative_error, use_exact=False):
    rf = prop.raw_formula
    player_index = None
    if rf.is_reward_operator and use_exact:
        raise ValueError("exact synthesis is not supported for reward properties")
    
    if not (rf.is_reward_operator or rf.is_probability_operator) and rf.is_game_formula:
        if use_exact:
            raise ValueError("exact synthesis is not supported for game properties")
        
        player_index = extract_player_index(rf)
        game_rf = rf
        rf = rf.subformula
        prop = stormpy.Property("", rf)
    assert rf.has_bound != rf.has_optimality_type, \
        "optimizing formula contains a bound or a comparison formula does not"
    if rf.has_bound:
        prop = Property(prop, use_exact)
    else:
        prop = OptimalityProperty(prop, relative_error, use_exact)

    if player_index is not None:
        prop.game_optimizing_player = player_index
        prop.game_formula = game_rf
        alt_formula_str = f"<<{prop.game_optimizing_player}>> " + prop.formula_alt.__str__()
        formulas = stormpy.parse_properties(alt_formula_str)
        prop.game_formula_alt = formulas[0].raw_formula

    return prop

def extract_player_index(formula):
    # TODO add support for multiple players in coalition
    string = formula.__str__()
    l_idx = string.index('<<')
    r_idx = string.index('>>')
    player_num = string[l_idx + len('<<') : r_idx]
    return int(player_num)

def construct_reward_property(reward_name, minimizing, target_label):
    direction = "min" if minimizing else "max"
    formula_str = 'R{"' + reward_name + '"}' + '{}=? [F "{}"]'.format(direction, target_label)
    formula = stormpy.parse_properties_without_context(formula_str)[0]
    optimality = OptimalityProperty(formula, 0)
    return optimality

class Property:
    ''' Wrapper over a stormpy property. '''

    # model checking environment (method & precision)
    environment = None
    # model checking precision
    model_checking_precision = 1e-4

    @classmethod
    def set_model_checking_precision(cls, precision):
        cls.model_checking_precision = precision
        payntbind.synthesis.set_precision_native(cls.environment.solver_environment.native_solver_environment, precision)
        payntbind.synthesis.set_precision_minmax(cls.environment.solver_environment.minmax_solver_environment, precision)

    @classmethod
    def initialize(cls, use_exact=False):
        cls.environment = stormpy.Environment()
        cls.set_model_checking_precision(cls.model_checking_precision)

        se = cls.environment.solver_environment
        # se.set_linear_equation_solver_type(stormpy.EquationSolverType.native)
        # se.set_linear_equation_solver_type(stormpy.EquationSolverType.gmmxx)
        # se.set_linear_equation_solver_type(stormpy.EquationSolverType.eigen)
        se.set_linear_equation_solver_type(stormpy.EquationSolverType.topological)

        if use_exact:
            se.minmax_solver_environment.method = stormpy.MinMaxMethod.policy_iteration
        else:
            se.minmax_solver_environment.method = stormpy.MinMaxMethod.optimistic_value_iteration

    @classmethod
    def model_check(cls, model, formula):
        return stormpy.model_checking(model, formula, extract_scheduler=True, environment=cls.environment)

    @classmethod
    def compute_expected_visits(cls, model):
        result = stormpy.compute_expected_number_of_visits(cls.environment, model)
        values = list(result.get_values())
        return values

    @staticmethod
    def above_model_checking_precision(a, b):
        if isinstance(a, stormpy.Rational):
            return True
        return abs(a-b) > Property.model_checking_precision


    def __init__(self, prop, use_exact=False):
        self.property = prop
        rf = prop.raw_formula

        self.game_optimizing_player = None # player index for game properties
        self.game_formula = None

        self.use_exact = use_exact

        # use comparison type to deduce optimizing direction
        comparison_type = rf.comparison_type
        self.minimizing = comparison_type in [stormpy.ComparisonType.LESS, stormpy.ComparisonType.LEQ]
        self.op = {
            stormpy.ComparisonType.LESS:    operator.lt,
            stormpy.ComparisonType.LEQ:     operator.le,
            stormpy.ComparisonType.GREATER: operator.gt,
            stormpy.ComparisonType.GEQ:     operator.ge
        }[comparison_type]

        # set threshold
        if use_exact:
            self.threshold = rf.threshold_expr.evaluate_as_rational()
            self.threshold_plus_precision = self.threshold
        else:
            self.threshold = rf.threshold_expr.evaluate_as_double()
            if self.minimizing:
                self.threshold_plus_precision = self.threshold + Property.model_checking_precision
            else:
                self.threshold_plus_precision = self.threshold - Property.model_checking_precision

        # construct quantitative formula (without bound) for explicit model checking
        # set optimality type
        self.formula = rf.clone()
        self.formula.remove_bound()
        if self.minimizing:
            self.formula.set_optimality_type(stormpy.OptimizationDirection.Minimize)
        else:
            self.formula.set_optimality_type(stormpy.OptimizationDirection.Maximize)
        self.formula_alt = Property.alt_formula(self.formula)

    @staticmethod
    def alt_formula(formula):
        '''
        :return formula with the opposite optimality type
        '''
        formula_alt = formula.clone()
        optimality_type = formula.optimality_type
        if optimality_type == stormpy.OptimizationDirection.Minimize:
            optimality_type = stormpy.OptimizationDirection.Maximize
        else:
            optimality_type = stormpy.OptimizationDirection.Minimize
        formula_alt.set_optimality_type(optimality_type)
        return formula_alt

    def __str__(self):
        return str(self.property.raw_formula)

    @property
    def reward(self):
        return self.formula.is_reward_operator

    @property
    def is_discounted_reward(self):
        # TODO add discounted reward as a type to Stormpy formula
        # return self.formula.is_reward_operator and self.formula.subformula.is_discounted_total_reward_formula
        return self.formula.is_reward_operator and "discount" in str(self.formula.subformula)

    @property
    def maximizing(self):
        return not self.minimizing

    @property
    def is_until(self):
        return self.formula.subformula.is_until_formula

    @property
    def has_game_formula(self):
        return self.game_formula is not None

    def transform_until_to_eventually(self):
        if not self.is_until:
            return
        logger.info("converting until formula to eventually...")
        formula = payntbind.synthesis.transform_until_to_eventually(self.property.raw_formula)
        prop = stormpy.Property("", formula)
        self.__init__(prop, self.use_exact)

    def property_copy(self):
        return stormpy.Property("", self.property.raw_formula.clone())

    def copy(self):
        return Property(self.property_copy())

    def result_valid(self, value):
        return not self.reward or value != math.inf

    def satisfies_threshold(self, value):
        return self.result_valid(value) and self.op(value, self.threshold)

    def satisfies_threshold_within_precision(self, value):
        return self.result_valid(value) and self.op(value, self.threshold_plus_precision)

    @property
    def can_be_improved(self):
        return False

    def negate(self):
        negated_formula = self.property.raw_formula.clone()
        negated_formula.comparison_type = {
            stormpy.ComparisonType.LESS:    stormpy.ComparisonType.GEQ,
            stormpy.ComparisonType.LEQ:     stormpy.ComparisonType.GREATER,
            stormpy.ComparisonType.GREATER: stormpy.ComparisonType.LEQ,
            stormpy.ComparisonType.GEQ:     stormpy.ComparisonType.LESS
        }[negated_formula.comparison_type]
        stormpy_property_negated = stormpy.Property("", negated_formula)
        property_negated = Property(stormpy_property_negated)
        return property_negated

    def get_target_label(self):
        target = self.formula.subformula.subformula
        if isinstance(target, stormpy.logic.AtomicLabelFormula):
            target_label = target.label
        elif isinstance(target, stormpy.logic.AtomicExpressionFormula):
            target_label = str(target)
        else:
            raise ValueError(f"unknown type of target expression {str(target)}, expected atomic label or atomic expression")
        return target_label

    def get_reward_name(self):
        assert self.reward
        return self.formula.reward_name

    def transform_to_optimality_formula(self, prism):
        direction = "min" if self.minimizing else "max"
        if self.reward:
            if isinstance(self.formula.subformula.subformula, stormpy.logic.AtomicLabelFormula):
                formula_str = f"R{{\"{self.get_reward_name()}\"}}{direction}=? [F \"{self.get_target_label()}\"]"
            else:
                formula_str = f"R{{\"{self.get_reward_name()}\"}}{direction}=? [F {self.get_target_label()}]"
        else:
            if isinstance(self.formula.subformula.subformula, stormpy.logic.AtomicLabelFormula):
                formula_str = f"P{direction}=? [F \"{self.get_target_label()}\"]"
            else:
                formula_str = f"P{direction}=? [F {self.get_target_label()}]"
        formula = stormpy.parse_properties_for_prism_program(formula_str, prism)[0]
        return formula



class OptimalityProperty(Property):
    '''
    Optimality property can remember current optimal value and adapt the
    corresponding threshold wrt epsilon.
    '''
    def __init__(self, prop, epsilon=0, use_exact=False):
        self.property = prop
        rf = prop.raw_formula

        self.game_optimizing_player = None # player index for game properties
        self.game_formula = None

        self.use_exact = use_exact

        # use comparison type to deduce optimizing direction
        if rf.optimality_type == stormpy.OptimizationDirection.Minimize:
            self.minimizing = True
            self.op = operator.lt
        else:
            self.minimizing = False
            self.op = operator.gt

        # construct quantitative formula (without bound) for explicit model checking
        self.formula = rf.clone()
        self.formula_alt = Property.alt_formula(self.formula)

        # additional optimality stuff
        self.optimum = None
        if use_exact:
            self.epsilon = stormpy.Rational(epsilon)
        else:
            self.epsilon = epsilon

        self.reset()


    def __str__(self):
        eps = f"[eps = {self.epsilon}]" if self.epsilon > 0 else ""
        return f"{str(self.formula)} {eps}"

    def copy(self):
        return OptimalityProperty(self.property_copy(), self.epsilon)

    def reset(self):
        self.optimum = None
        if self.minimizing:
            if self.use_exact:
                self.threshold = stormpy.Rational(2) # TODO: does not work for rewards
            else:
                self.threshold = math.inf
        else:
            if self.use_exact:
                self.threshold = stormpy.Rational(-1) # TODO: does not work for rewards
            else:
                self.threshold = -math.inf

    def meets_op(self, a, b):
        ''' For optimality objective, we want to accept improvements above model checking precision. '''
        return b is None or (Property.above_model_checking_precision(a,b) and self.op(a,b))

    def satisfies_threshold(self, value):
        return self.result_valid(value) and self.meets_op(value, self.threshold)

    def improves_optimum(self, value):
        return self.result_valid(value) and self.meets_op(value, self.optimum)

    def update_optimum(self, optimum):
        self.optimum = optimum
        if self.minimizing:
            self.threshold = optimum * (1 - self.epsilon)
        else:
            self.threshold = optimum * (1 + self.epsilon)

    def suboptimal_value(self):
        assert self.optimum is not None
        if self.minimizing:
            return self.optimum * (1 + self.model_checking_precision)
        else:
            return self.optimum * (1 - self.model_checking_precision)

    def transform_until_to_eventually(self):
        if not self.is_until:
            return
        logger.info("converting until formula to eventually...")
        formula = payntbind.synthesis.transform_until_to_eventually(self.property.raw_formula)
        prop = stormpy.Property("", formula)
        self.__init__(prop, self.epsilon, self.use_exact)

    @property
    def can_be_improved(self):
        return not( not self.reward and self.minimizing and self.threshold == 0 )

    def negate(self):
        negated_formula = self.property.raw_formula.clone()
        negate_optimality_type = {
            stormpy.OptimizationDirection.Minimize:    stormpy.OptimizationDirection.Maximize,
            stormpy.OptimizationDirection.Maximize:    stormpy.OptimizationDirection.Minimize
        }[negated_formula.optimality_type]
        negated_formula.set_optimality_type(negate_optimality_type)
        stormpy_property_negated = stormpy.Property("", negated_formula)
        property_negated = OptimalityProperty(stormpy_property_negated,self.epsilon)
        return property_negated


class Specification:

    def __init__(self, properties):
        self.constraints = []
        self.optimality = None

        # sort the properties
        optimalities = []
        for p in properties:
            if type(p) == Property:
                self.constraints.append(p)
            if type(p) == OptimalityProperty:
                optimalities.append(p)
        assert len(optimalities) <=1, "multiple optimality objectives were specified"
        if optimalities:
            self.optimality = optimalities[0]

    def __str__(self):
        s = ""
        if self.constraints:
            s += "constraints: " + ",".join([str(c) for c in self.constraints]) + "; "
        if self.optimality is not None:
            s += "optimality: " + str(self.optimality)
        return s

    def copy(self):
        properties = [p.copy() for p in self.all_properties()]
        return Specification(properties)

    def reset(self):
        if self.optimality is not None:
            self.optimality.reset()

    @property
    def has_optimality(self):
        return self.optimality is not None

    @property
    def num_properties(self):
        return len(self.constraints) + (1 if self.has_optimality else 0)

    @property
    def is_single_property(self):
        return self.num_properties == 1

    def all_properties(self):
        properties = [c for c in self.constraints]
        if self.has_optimality:
            properties += [self.optimality]
        return properties

    def all_constraint_indices(self):
        return range(len(self.constraints))

    def stormpy_properties(self):
        return [p.property for p in self.all_properties()]

    def stormpy_formulae(self):
        return [p.formula for p in self.all_properties()]

    def contains_until_properties(self):
        return any([p.is_until for p in self.all_properties()])

    def transform_until_to_eventually(self):
        for p in self.all_properties():
            p.transform_until_to_eventually()


    def check(self):
        # TODO
        pass

    def can_be_improved(self):
        return any(prop.can_be_improved for prop in self.all_properties())

    @property
    def contains_maximizing_reward_properties(self):
        return any([c.reward and not c.minimizing for c in self.all_properties()])

    def negate(self):
        properties_negated = [p.negate() for p in self.all_properties()]
        return Specification(properties_negated)




from ._info import *
from . import _config


def storm_version() -> str:
    """
    Get Storm version.

    :return: Storm version
    :rtype: str
    """
    return _config.STORM_VERSION


def storm_from_system() -> bool:
    """
    Does payntbind rely on a (pre-existing) system installation of Storm.
    System versions include anything on the system
    that was not fetched during the build of payntbind.

    :return: True, if  payntbind linked against a preexisting Storm installation.
    """
    return _config.STORM_FROM_SYSTEM


def storm_development_version() -> bool:
    """
    Is true, if the version of Storm was marked as modified since release.
    :return: True if a development version
    """
    return _config.STORM_DEVELOPER_VERSION


def storm_origin_info() -> [str | None, str | None]:
    """
    Information about the source of Storm, in particular the repo path and the repo tag.

    If Storm was already on the system during installation and this preexisting version was used,
     these values are expected to be None.
    If Storm was installed (locally) as part of the installation process, these values are expected to be not None.
    Then:
      - The repo path usually takes the form of a url
      - The tag refers to a tag on the repo.
    A noteworthy exception is when the fetch was from a local source dir.
    In that case, this source dir is included, and the tag reads "__local-source-dir__"

    :return: A pair with the repo path and the repo tag.
    """
    return (_config.STORM_ORIGIN_REPO, _config.STORM_ORIGIN_TAG)


def storm_directory() -> str | None:
    """
    Return the Storm directory which is used by payntbind.

    If a prexisting installation of Storm was used, then the path to this directory is returned.
    If Storm was installed during the installation process, value None is returned.

    :return: Storm directory.
    """
    return _config.STORM_DIRECTORY


import payntbind

import logging
logger = logging.getLogger(__name__)


class ConflictGeneratorDtmc():

    def __init__(self, quotient):
        self.quotient = quotient
        self.counterexample_generator = None

    @property
    def name(self):
        return "(DTMC)"

    def initialize(self):
        state_to_holes_bv = self.quotient.coloring.getStateToHoles().copy()
        state_to_holes = []
        for state,holes_bv in enumerate(state_to_holes_bv):
            holes = set([hole for hole in holes_bv])
            state_to_holes.append(holes)
        formulae = self.quotient.specification.stormpy_formulae()
        self.counterexample_generator = payntbind.synthesis.CounterexampleGenerator(
            self.quotient.quotient_mdp, self.quotient.family.num_holes,
            state_to_holes, formulae
        )


    def prepare_model(self, model):
        self.counterexample_generator.prepare_dtmc(model.model, model.quotient_state_map)

    def construct_conflicts(self, family, assignment, dtmc, conflict_requests):
        
        self.prepare_model(dtmc)
        
        conflicts = []
        for request in conflict_requests:
            index,prop,family_result = request

            threshold = prop.threshold

            bounds = None
            scheduler_selection = None
            if family_result is not None:
                bounds = family_result.primary.result

            conflict = self.counterexample_generator.construct_conflict(index, threshold, bounds, family.mdp.quotient_state_map)
            conflicts.append(conflict)
        
        return conflicts



import paynt.synthesizer.statistic
import paynt.utils.timer

import logging
logger = logging.getLogger(__name__)


class FamilyEvaluation:
    '''Result associated with a family after its evaluation. '''
    def __init__(self, family, value, sat, policy):
        self.family = family
        self.value = value
        self.sat = sat
        self.policy = policy


class Synthesizer:

    # base filename (i.e. without extension) to export synthesis result
    export_synthesis_filename_base = None

    @staticmethod
    def choose_synthesizer(quotient, method, fsc_synthesis=False, storm_control=None):

        # hiding imports here to avoid mutual top-level imports
        import paynt.quotient.mdp
        import paynt.quotient.pomdp
        import paynt.quotient.decpomdp
        import paynt.quotient.mdp_family
        import paynt.quotient.posmg
        import paynt.synthesizer.synthesizer_onebyone
        import paynt.synthesizer.synthesizer_ar
        import paynt.synthesizer.synthesizer_cegis
        import paynt.synthesizer.synthesizer_hybrid
        import paynt.synthesizer.synthesizer_multicore_ar
        import paynt.synthesizer.synthesizer_pomdp
        import paynt.synthesizer.synthesizer_decpomdp
        import paynt.synthesizer.synthesizer_posmg
        import paynt.synthesizer.policy_tree
        import paynt.synthesizer.decision_tree

        if isinstance(quotient, paynt.quotient.pomdp_family.PomdpFamilyQuotient):
            logger.info("nothing to do with the POMDP sketch, aborting...")
            exit(0)
        if isinstance(quotient, paynt.quotient.mdp.MdpQuotient):
            return paynt.synthesizer.decision_tree.SynthesizerDecisionTree(quotient)
        # FSC synthesis for POMDPs
        if isinstance(quotient, paynt.quotient.pomdp.PomdpQuotient) and fsc_synthesis:
            return paynt.synthesizer.synthesizer_pomdp.SynthesizerPomdp(quotient, method, storm_control)
        # FSC synthesis for Dec-POMDPs
        if isinstance(quotient, paynt.quotient.decpomdp.DecPomdpQuotient) and fsc_synthesis:
            return paynt.synthesizer.synthesizer_decpomdp.SynthesizerDecPomdp(quotient)
        # Policy Tree synthesis for family of MDPs
        if isinstance(quotient, paynt.quotient.mdp_family.MdpFamilyQuotient):
            if method == "onebyone":
                return paynt.synthesizer.synthesizer_onebyone.SynthesizerOneByOne(quotient)
            else:
                return paynt.synthesizer.policy_tree.SynthesizerPolicyTree(quotient)
        # FSC synthesis for POSMGs
        if isinstance(quotient, paynt.quotient.posmg.PosmgQuotient) and fsc_synthesis:
            return paynt.synthesizer.synthesizer_posmg.SynthesizerPosmg(quotient)

        # synthesis engines
        if method == "onebyone":
            return paynt.synthesizer.synthesizer_onebyone.SynthesizerOneByOne(quotient)
        if method == "ar":
            return paynt.synthesizer.synthesizer_ar.SynthesizerAR(quotient)
        if method == "cegis":
            return paynt.synthesizer.synthesizer_cegis.SynthesizerCEGIS(quotient)
        if method == "hybrid":
            return paynt.synthesizer.synthesizer_hybrid.SynthesizerHybrid(quotient)
        if method == "ar_multicore":
            return paynt.synthesizer.synthesizer_multicore_ar.SynthesizerMultiCoreAR(quotient)
        raise ValueError("invalid method name")


    def __init__(self, quotient):
        self.quotient = quotient
        self.stat = None
        self.synthesis_timer = None
        self.explored = None
        self.best_assignment = None
        self.best_assignment_value = None

    @property
    def method_name(self):
        ''' to be overridden '''
        pass

    def time_limit_reached(self):
        if (self.synthesis_timer is not None and self.synthesis_timer.time_limit_reached()) or \
            paynt.utils.timer.GlobalTimer.time_limit_reached():
            logger.info("time limit reached, aborting...")
            return True
        return False

    def memory_limit_reached(self):
        if paynt.utils.timer.GlobalMemoryLimit.limit_reached():
            logger.info("memory limit reached, aborting...")
            return True
        return False

    def resource_limit_reached(self):
        return self.time_limit_reached() or self.memory_limit_reached()

    def set_optimality_threshold(self, optimum_threshold):
        if self.quotient.specification.has_optimality and optimum_threshold is not None:
            self.quotient.specification.optimality.update_optimum(optimum_threshold)
            logger.debug(f"optimality threshold set to {optimum_threshold}")

    def explore(self, family):
        self.explored += family.size

    def evaluate_all(self, family, prop, keep_value_only=False):
        ''' to be overridden '''
        pass

    def export_evaluation_result(self, evaluations, export_filename_base):
        ''' to be overridden '''
        pass

    def evaluate(self, family=None, prop=None, keep_value_only=False, print_stats=True):
        '''
        Evaluate each member of the family wrt the given property.
        :param family if None, then the design space of the quotient will be used
        :param prop if None, then the default property of the quotient will be used
            (assuming single-property specification)
        :param keep_value_only if True, only value will be associated with the family
        :param print_stats if True, synthesis statistic will be printed
        :param export_filename_base base filename used to export the evaluation results
        :returns a list of (family,evaluation) pairs
        '''
        if family is None:
            family = self.quotient.family
        if prop is None:
            prop = self.quotient.get_property()

        self.stat = paynt.synthesizer.statistic.Statistic(self)
        self.explored = 0
        logger.info("evaluation initiated, design space: {}".format(family.size))
        self.stat.start(family)
        evaluations = self.evaluate_all(family, prop, keep_value_only)
        self.stat.finished_evaluation(evaluations)
        logger.info("evaluation finished")

        if self.export_synthesis_filename_base is not None:
            self.export_evaluation_result(evaluations, self.export_synthesis_filename_base)

        if print_stats:
            self.stat.print()

        return evaluations


    def synthesize_one(self, family):
        ''' to be overridden '''
        pass

    def synthesize(
        self, family=None, optimum_threshold=None, keep_optimum=False, return_all=False, print_stats=True, timeout=None
    ):
        '''
        :param family family of assignment to search in
        :param families alternatively, a list of families can be given
        :param optimum_threshold known bound on the optimum value
        :param keep_optimum if True, the optimality specification will not be reset upon finish
        :param return_all if True and the synthesis returns a family, all assignments will be returned instead of an
            arbitrary one
        :param print_stats if True, synthesis stats will be printed upon completion
        :param timeout synthesis time limit, seconds
        '''
        if family is None:
            family = self.quotient.family
        if family.constraint_indices is None:
            family.constraint_indices = list(range(len(self.quotient.specification.constraints)))

        self.set_optimality_threshold(optimum_threshold)
        self.synthesis_timer = paynt.utils.timer.Timer(timeout)
        self.synthesis_timer.start()
        self.stat = paynt.synthesizer.statistic.Statistic(self)
        self.explored = 0
        self.stat.start(family)
        self.synthesize_one(family)
        if self.best_assignment is not None and self.best_assignment.size > 1 and not return_all:
            self.best_assignment = self.best_assignment.pick_any()
        self.stat.finished_synthesis()
        if self.best_assignment is not None:
            logger.info("printing synthesized assignment below:")
            logger.info(self.best_assignment)

        if self.best_assignment is not None and self.best_assignment.size == 1:
            dtmc = self.quotient.build_assignment(self.best_assignment)
            result = dtmc.check_specification(self.quotient.specification)
            logger.info(f"double-checking specification satisfiability: {result}")

        if print_stats:
            self.stat.print()

        assignment = self.best_assignment
        if not keep_optimum:
            self.best_assignment = None
            self.best_assignment_value = None
            self.quotient.specification.reset()

        return assignment


    def run(self, optimum_threshold=None):
        return self.synthesize(optimum_threshold=optimum_threshold)



import stormpy
import payntbind

import paynt.family.family
import paynt.verification.property
import paynt.parser.jani
import paynt.models.model_builder

import os
import re
import uuid

import logging
logger = logging.getLogger(__name__)


class PrismParser:

    @classmethod
    def read_prism(cls, sketch_path, properties_path, relative_error, use_exact=False):

        # parse the program
        prism, hole_definitions = PrismParser.load_sketch_prism(sketch_path)
        expression_parser = stormpy.storage.ExpressionParser(prism.expression_manager)
        expression_parser.set_identifier_mapping(dict())
        prism_model_type = {
            stormpy.storage.PrismModelType.DTMC:"DTMC",
            stormpy.storage.PrismModelType.MDP:"MDP",
            stormpy.storage.PrismModelType.POMDP:"POMDP"
        }[prism.model_type]
        logger.debug("PRISM model type: " + prism_model_type)

        # parse constants
        constant_map = None

        # parse hole definitions
        hole_expressions = None
        family = None
        if len(hole_definitions) > 0:
            logger.info("processing hole definitions...")
            prism, hole_expressions, family = PrismParser.parse_holes(prism, expression_parser, hole_definitions)
        prism = prism.label_unlabelled_commands({})

        specification = PrismParser.parse_specification(properties_path, relative_error, prism, use_exact=use_exact)

        # construct the quotient
        coloring = None
        jani_unfolder = None
        obs_evaluator = None
        if family is not None:
            assert prism_model_type in ["DTMC","MDP","POMDP"], "hole detected, but the program is neither DTMC nor (PO)MDP"
            # unfold hole options via Jani
            jani_unfolder = paynt.parser.jani.JaniUnfolder(prism, hole_expressions, specification, family)
            specification = jani_unfolder.specification
            quotient_mdp = jani_unfolder.quotient_mdp
            coloring = payntbind.synthesis.Coloring(family.family, quotient_mdp.nondeterministic_choice_indices, jani_unfolder.choice_to_hole_options)
            if prism.model_type == stormpy.storage.PrismModelType.POMDP:
                obs_evaluator = payntbind.synthesis.ObservationEvaluator(prism, quotient_mdp)
            quotient_mdp = payntbind.synthesis.addChoiceLabelsFromJani(quotient_mdp)
        else:
            quotient_mdp = paynt.models.model_builder.ModelBuilder.from_prism(prism, specification, use_exact)

        return prism, quotient_mdp, specification, family, coloring, jani_unfolder, obs_evaluator

    
    @classmethod
    def load_sketch_prism(cls, sketch_path):
        # read lines
        with open(sketch_path) as f:
            sketch_lines = f.readlines()

        # replace hole definitions with constants
        hole_re_brace = re.compile(r'^\s*hole\s+(.*?)\s+(.*?)\s+in\s+\{(.*?)\}\s*;')
        # hole_re_bracket = re.compile(r'^\s*hole\s+(.*?)\s+(.*?)\s+in\s+[(.*?)]\s+;')
        sketch_output = []
        hole_definitions = []
        observables_line = False
        for line in sketch_lines:

            # Treat observation definition via "observables" keyword
            if line.startswith("observables"):
                observables_line = True
                line = line.split('//')[0].strip()  # remove comments
                if len(line) > len("observables"):
                    line = line[len("observables"):].strip()
                else:
                    continue
            if observables_line:
                if line.startswith("endobservables"):
                    observables_line = False
                    continue
                line = line.split('//')[0].strip()  # remove comments
                observables = line.strip().split(",")
                observables = [obs.strip() for obs in observables]
                for obs in observables:
                    sketch_output.append(f"observable \"{obs}\" = {obs};\n")
                continue
            

            match = hole_re_brace.search(line)
            if match is None:
                sketch_output.append(line)
                continue
            hole_type = match.group(1)
            hole_name = match.group(2)
            hole_options = match.group(3).replace(" ", "")
            hole_definitions.append( (hole_name,hole_type,hole_options) )
            sketch_output.append(f"const {hole_type} {hole_name};\n")
            sketch_output.append(f"const {hole_type} {hole_name}_MIN;\n")
            sketch_output.append(f"const {hole_type} {hole_name}_MAX;\n")


        # store modified sketch to a temporary file
        tmp_path = sketch_path + str(uuid.uuid4())
        with open(tmp_path, 'w') as f:
            for line in sketch_output:
                print(line, end="", file=f)

        # try to parse temporary sketch and then delete it
        try:
            prism = stormpy.parse_prism_program(tmp_path, prism_compat=True)
            os.remove(tmp_path)
        except:
            os.remove(tmp_path)
            raise SyntaxError

        return prism, hole_definitions


    @classmethod
    def parse_holes(cls, prism, expression_parser, hole_definitions):

        # parse hole definitions
        family = paynt.family.family.Family()
        hole_expressions = []
        hole_min = []
        hole_max = []
        for hole_name,hole_type,hole_options in hole_definitions:
            if ".." in hole_options:
                assert hole_type == "int" or hole_type == "double", "cannot use range-based definitions for non-integer of non-double hole types"
                if hole_type == "double":
                    assert ":" in hole_options, "using range-based definition for double requires specifying the increment step range_start..range_end:step"
                    range_start = float(hole_options[0:hole_options.find('..')])
                    range_end = float(hole_options[hole_options.find('..')+2:hole_options.find(':')].strip())
                    increment_string = hole_options.split(':')[-1].strip()
                    increment = float(increment_string)
                    increment_decimal_precision = len(increment_string.split('.')[-1])
                    hole_min.append(range_start)
                    hole_max.append(range_end)
                    steps = (range_end - range_start) / increment
                    options = ["{:.{n}f}".format(range_start + x * increment, n=increment_decimal_precision) for x in range(int(round(steps)+1))]
                    if float(options[-1]) > range_end:
                        options = options[:-1]
                else:
                    if ":" in hole_options:
                        range_start = int(hole_options[0:hole_options.find('..')])
                        range_end = int(hole_options[hole_options.find('..')+2:hole_options.find(':')].strip())
                        increment = int(hole_options.split(':')[-1].strip())
                    else:
                        options = hole_options.split("..")
                        range_start = int(options[0])
                        range_end = int(options[1])
                        increment = 1
                    hole_min.append(range_start)
                    hole_max.append(range_end)
                    options = [str(o) for o in range(range_start,range_end+1, increment)]
            else:
                options = hole_options.split(",")
                if hole_type == "int":
                    options_numerical = [int(o) for o in options]
                else:
                    options_numerical = [float(o) for o in options]
                hole_min.append(min(options_numerical))
                hole_max.append(max(options_numerical))
            expressions = [expression_parser.parse(o) for o in options]
            hole_expressions.append(expressions)

            option_labels = [str(e) for e in expressions]
            family.add_hole(hole_name, option_labels)

        # substitute constants used as min/max values of holes
        hole_range_definitions = {}
        for hole in range(family.num_holes):
            hole_name = family.hole_name(hole)
            var_min = prism.get_constant(f"{hole_name}_MIN").expression_variable
            hole_range_definitions[var_min] = expression_parser.parse(str(hole_min[hole]))
            var_max = prism.get_constant(f"{hole_name}_MAX").expression_variable
            hole_range_definitions[var_max] = expression_parser.parse(str(hole_max[hole]))
        prism = prism.define_constants(hole_range_definitions)

        # check that all undefined constants are indeed the holes
        hole_names = [family.hole_name(hole) for hole in range(family.num_holes)]
        for c in prism.constants:
            if not c.defined:
                assert c.name in hole_names, f"constant {c.name} was not specified"

        return prism, hole_expressions, family

 
    @classmethod
    def parse_property(cls, line, prism=None):
        '''
        Parse a line containing a single PCTL property.
        @return the property or None if no property was detected
        '''
        if prism is not None:
            props = stormpy.parse_properties_for_prism_program(line, prism)
        else:
            props = stormpy.parse_properties_without_context(line)
        if len(props) == 0:
            return None
        if len(props)>1:
            logger.warning("multiple properties detected on one line, dropping all but the first one")
        return props[0]

    @classmethod
    def parse_specification(cls, properties_path, relative_error=0, prism=None, use_exact=False):
        '''
        Expecting one property per line. The line may be terminated with a semicolon.
        Empty lines or comments are allowed.
        '''
        if not os.path.isfile(properties_path):
            raise ValueError(f"the properties file {properties_path} does not exist")
        logger.info(f"loading properties from {properties_path} ...")

        lines = ""
        with open(properties_path) as file:
            lines = [line for line in file]
        
        properties = []

        for line in lines:
            formula = PrismParser.parse_property(line,prism)
            if formula is None:
                continue
            prop = paynt.verification.property.construct_property(formula, relative_error, use_exact)
            properties.append(prop)

        specification = paynt.verification.property.Specification(properties)
        logger.info(f"found the following specification: {specification}")
        return specification



import stormpy.storage

import paynt.utils.timer
import paynt.synthesizer.synthesizer
import paynt.models.models

import math

import logging
logger = logging.getLogger(__name__)

# zero approximation to avoid zero division exception
APPROX_ZERO = 0.000001

def safe_division(dividend, divisor):
    """Safe division of dividend by operand
    :param number dividend: upper operand of the division
    :param number divisor: lower operand of the division, may be zero
    :returns safe value after division of approximated zero
    """
    try:
        return dividend / divisor
    except (ZeroDivisionError, ValueError):
        return dividend / APPROX_ZERO

class Statistic:
    """General computation stats."""

    # parameters
    status_period_seconds = 3
    synthesis_timer_total = paynt.utils.timer.Timer()
    
    def __init__(self, synthesizer):
        
        self.synthesizer = synthesizer
        self.quotient = self.synthesizer.quotient

        self.iterations_dtmc = None
        self.acc_size_dtmc = 0
        self.avg_size_dtmc = 0

        self.iterations_mdp = None
        self.acc_size_mdp = 0
        self.avg_size_mdp = 0

        self.iterations_game = None
        self.acc_size_game = 0
        self.avg_size_game = 0

        self.synthesized_assignment = None
        self.job_type = None

        # MDP family
        self.num_mdps_total = None
        self.num_mdps_sat = None
        self.num_tree_nodes = None
        self.num_tree_nodes_merged = None
        self.num_policies = None
        self.num_policies_merged = None

        self.family_size = None
        self.synthesis_timer = paynt.utils.timer.Timer()
        self.status_horizon = Statistic.status_period_seconds


    def start(self, family):
        logger.info("synthesis initiated, design space: {}".format(family.size_or_order))
        self.family_size = family.size
        self.synthesis_timer.start()
        if not self.synthesis_timer_total.running:
            self.synthesis_timer_total.start()
    
    def iteration(self, model):
        ''' Identify the type of the model and count corresponding iteration. '''
        if isinstance(model, paynt.models.models.Mdp):
            model = model.model
        if type(model) in [stormpy.storage.SparseDtmc, stormpy.storage.SparseExactDtmc]:
            self.iteration_dtmc(model.nr_states)
        elif type(model) in [stormpy.storage.SparseMdp, stormpy.storage.SparseExactMdp]:
            self.iteration_mdp(model.nr_states)
        else:
            logger.debug(f"unknown model type {type(model)}")

    def iteration_dtmc(self, size_dtmc):
        if self.iterations_dtmc is None:
            self.iterations_dtmc = 0
        self.iterations_dtmc += 1
        self.acc_size_dtmc += size_dtmc
        self.print_status()

    def iteration_mdp(self, size_mdp):
        if self.iterations_mdp is None:
            self.iterations_mdp = 0
        self.iterations_mdp += 1
        self.acc_size_mdp += size_mdp
        self.print_status()

    def iteration_game(self, size_game):
        if self.iterations_game is None:
            self.iterations_game = 0
        self.iterations_game += 1
        self.acc_size_game += size_game
        self.print_status()

    def new_fsc_found(self, value, assignment, size):
        time_elapsed = round(self.synthesis_timer_total.read(),1)
        # print(f'new opt: {value}')
        # print(f'new opt: {value}, elapsed {time_elapsed}s')
        # print(f'-----------PAYNT----------- \
              # \nValue = {value} | Time elapsed = {time_elapsed}s | FSC size = {size}\nFSC = {assignment}\n', flush=True)

    
    def status(self):
        ret_str = "> "
        fraction_explored = self.synthesizer.explored / self.family_size
        time_estimate = safe_division(self.synthesis_timer.read(), fraction_explored)
        percentage_explored = int(fraction_explored * 100000) / 1000.0
        ret_str += f"progress {percentage_explored}%"
        
        time_elapsed = int(self.synthesis_timer.read())
        ret_str += f", elapsed {time_elapsed} s"
        time_estimate = int(time_estimate)
        ret_str += f", estimated {time_estimate} s"
        time_estimate_hours = math.floor(time_estimate/3600)
        time_estimate_days = math.floor(time_estimate_hours/24)
        time_estimate_years = math.floor(time_estimate_days/365)
        if time_estimate_years > 0:
            s_ending = "s" if time_estimate_years > 1 else ""
            ret_str += f" ({time_estimate_years} year{s_ending})"
        elif time_estimate_days > 1:
            s_ending = "s" if time_estimate_days > 1 else ""
            ret_str += f" ({time_estimate_days} day{s_ending})"
        elif time_estimate_hours > 1:
            s_ending = "s" if time_estimate_hours > 1 else ""
            ret_str += f" ({time_estimate_hours} hour{s_ending})"

        iters = []
        if self.iterations_game is not None:
            iters += [f"game: {self.iterations_game}"]
        if self.iterations_mdp is not None:
            iters += [f"MDP: {self.iterations_mdp}"]
        if self.iterations_dtmc is not None:
            iters += [f"DTMC: {self.iterations_dtmc}"]
        ret_str += ", iters = {" + ", ".join(iters) + "}"
        # ret_str += f", pres = {self.synthesizer.num_preserved}"
        
        spec = self.quotient.specification
        if spec.has_optimality:
            opt = self.synthesizer.best_assignment_value
            if opt is None:
                opt = spec.optimality.optimum
            if opt is not None:
                if not isinstance(opt, stormpy.Rational):
                    opt = round(opt, 4)
                ret_str += f", opt = {opt}"
        return ret_str


    def print_status(self):
        if not self.synthesis_timer.read() > self.status_horizon:
            return
        print(self.status(), flush=True)
        self.status_horizon = self.synthesis_timer.read() + Statistic.status_period_seconds


    def finished_synthesis(self):
        self.job_type = "synthesis"
        self.synthesis_timer.stop()
        self.synthesized_assignment = self.synthesizer.best_assignment

    def finished_evaluation(self, evaluations):
        self.job_type = "evaluation"
        self.synthesis_timer.stop()
        self.evaluations = evaluations
        

    def get_summary_specification(self):
        spec = self.quotient.specification
        specification = ""
        if len(spec.constraints) > 0:
            specification += "\n".join([f"constraint {i + 1}: {str(f)}" for i,f in enumerate(spec.constraints)]) + "\n"
        if spec.has_optimality:
            specification += f"optimality objective: {str(spec.optimality)}\n"
        return specification

    def get_summary_iterations(self):
        iterations = ""
        if self.iterations_game is not None:
            avg_size = round(safe_division(self.acc_size_game, self.iterations_game))
            type_stats = f"Game stats: avg game size: {avg_size}, iterations: {self.iterations_game}" 
            iterations += f"{type_stats}\n"

        if self.iterations_mdp is not None:
            avg_size = round(safe_division(self.acc_size_mdp, self.iterations_mdp))
            type_stats = f"MDP stats: avg MDP size: {avg_size}, iterations: {self.iterations_mdp}"
            iterations += f"{type_stats}\n"

        if self.iterations_dtmc is not None:
            avg_size = round(safe_division(self.acc_size_dtmc, self.iterations_dtmc))
            type_stats = f"DTMC stats: avg DTMC size: {avg_size}, iterations: {self.iterations_dtmc}"
            iterations += f"{type_stats}\n"
        return iterations

    def get_summary_synthesis(self):
        spec = self.quotient.specification
        if spec.has_optimality and spec.optimality.optimum is not None:
            if isinstance(spec.optimality.optimum, stormpy.Rational):
                optimum = spec.optimality.optimum
            else:
                optimum = round(spec.optimality.optimum, 6)
            return f"optimum: {optimum}"
        else:
            feasible = "yes" if self.synthesized_assignment is not None else "no"
            return f"feasible: {feasible}"

    def get_summary_evaluation(self):
        if not self.evaluations or not isinstance(self.evaluations[0], paynt.synthesizer.synthesizer.FamilyEvaluation):
            return ""
        members_sat = sum( [evaluation.family.size for evaluation in self.evaluations if evaluation.sat ])
        members_total = self.quotient.family.size
        members_sat_percentage = int(round(members_sat/members_total*100,0))
        return f"satisfied {members_sat}/{members_total} members ({members_sat_percentage}%)"

    
    def get_summary(self):
        specification = self.get_summary_specification()

        fraction_explored = int((self.synthesizer.explored / self.family_size) * 100)
        explored = f"explored: {fraction_explored} %"

        quotient_states = self.quotient.quotient_mdp.nr_states
        quotient_actions = self.quotient.quotient_mdp.nr_choices
        design_space = f"number of holes: {self.quotient.family.num_holes}, family size: {self.quotient.family.size_or_order}, quotient: {quotient_states} states / {quotient_actions} actions"
        timing = f"method: {self.synthesizer.method_name}, synthesis time: {round(self.synthesis_timer.time, 2)} s"

        iterations = self.get_summary_iterations()
        
        if self.job_type == "synthesis":
            result = self.get_summary_synthesis()
        else:
            result = self.get_summary_evaluation()

        sep = "--------------------\n"
        summary = f"{sep}"\
                f"Synthesis summary:\n" \
                f"{specification}\n{timing}\n{design_space}\n{explored}\n" \
                f"{iterations}\n{result}\n"\
                f"{sep}"
        return summary
    
    def print(self):    
        print(self.get_summary(),end="")


    def print_mdp_family_table_entries(self):
        model_info = "model info:\t"
        model_info += "\t".join(["states","choices","MDPs","states*MDPs","SAT MDPs","SAT %",])
        print(model_info)
        # print("\t\t",end="")
        print(self.quotient.quotient_mdp.nr_states,end=" ")
        print(self.quotient.quotient_mdp.nr_choices,end=" ")
        print(self.num_mdps_total,end=" ")
        print(self.quotient.quotient_mdp.nr_states*self.num_mdps_total,end=" ")
        print(self.num_mdps_sat,end=" ")
        sat_by_total_percentage = round(self.num_mdps_sat/self.num_mdps_total*100,2)
        print(sat_by_total_percentage)

        headers = [
            "time","nodes","nodes (merged)","leaves","leaves (merged)","leaves (merged) / MDPs %",
            "policies","policies (merged)","policies (merged) / SAT %","pp time","pp time %",
            "game iters","MDP iters","iters/MDPs %"]
        synt_stats_header = "synthesis info:\t" + "\t".join(headers)
        print(synt_stats_header)

        # print("\t\t",end="")
        synthesis_time = int(self.synthesis_timer.time)
        print(synthesis_time,end=" ")
        print(self.num_nodes,end=" ")
        print(self.num_nodes_merged,end=" ")
        
        print(self.num_leaves,end=" ")
        print(self.num_leaves_merged,end=" ")
        leaves_by_mdps = round(self.num_leaves_merged/self.num_mdps_total*100,2)
        print(leaves_by_mdps,end=" ")

        print(self.num_policies,end=" ")
        print(self.num_policies_merged,end=" ")
        policies_by_sat = "N/A"
        if self.num_mdps_sat > 0:
            policies_by_sat = round(self.num_policies_merged/self.num_mdps_sat*100,2)
        print(policies_by_sat,end=" ")
        print(self.postprocessing_time,end=" ")
        postprocessing_time_percentage = round(self.postprocessing_time/synthesis_time*100,2)
        print(postprocessing_time_percentage,end=" ")

        print(self.iterations_game,end=" ")
        print(self.iterations_mdp,end=" ")
        iters_by_mdp = round((self.iterations_game+self.iterations_mdp)/self.num_mdps_total*100,2)
        print(iters_by_mdp)
        print()



import stormpy
import payntbind

import paynt.models.models
import paynt.quotient.quotient
import paynt.quotient.mdp_family
import paynt.verification.property


import logging
logger = logging.getLogger(__name__)

class SubPomdp:
    '''
    Simple container for a (sub-)POMDP created from the quotient.
    '''
    def __init__(self, model, quotient, quotient_state_map, quotient_choice_map):
        # the Stormpy POMDP
        self.model = model
        # POMDP family quotient from which this POMDP was constructed
        # self.quotient = quotient
        # for each state of the POMDP, a state in the quotient
        self.quotient_state_map = quotient_state_map
        # for each choice of the POMDP, a choice in the quotient
        self.quotient_choice_map = quotient_choice_map


class PomdpFamilyQuotient(paynt.quotient.mdp_family.MdpFamilyQuotient):

    def __init__(self, quotient_mdp, family, coloring, specification, obs_evaluator):
        super().__init__(quotient_mdp = quotient_mdp, family = family, coloring = coloring, specification = specification)

        # if memory was unfolded we need to update obs_evaluator
        if paynt.quotient.mdp_family.MdpFamilyQuotient.initial_memory_size > 1:
            prototype_states = list(self.memory_unfolder.state_prototype)
            state_to_obs_class = list(obs_evaluator.state_to_obs_class)
            new_obs_classes_map = [state_to_obs_class[prototype_states[state]] for state in range(self.quotient_mdp.nr_states)]
            obs_evaluator.state_to_obs_class = new_obs_classes_map

        self.obs_evaluator = obs_evaluator

        # for each observation, a list of actions (indices) available
        self.observation_to_actions = None
        # POMDP manager used for unfolding the memory model into the quotient POMDP
        self.fsc_unfolder = None

        # identify actions available at each observation
        self.observation_to_actions = [None] * self.num_observations
        state_to_observation = self.state_to_observation
        for state,available_actions in enumerate(self.state_to_actions):
            obs = state_to_observation[state]
            if self.observation_to_actions[obs] is not None:
                assert self.observation_to_actions[obs] == available_actions,\
                    f"two states in observation class {obs} differ in available actions"
                continue
            self.observation_to_actions[obs] = available_actions


    @property
    def num_observations(self):
        return self.obs_evaluator.num_obs_classes

    @property
    def state_to_observation(self):
        return self.obs_evaluator.state_to_obs_class

    def observation_is_trivial(self, obs):
        return len(self.observation_to_actions[obs])==1

    
    def build_pomdp(self, family):
        ''' Construct the sub-POMDP from the given hole assignment. '''
        assert family.size == 1, "expecting family of size 1"
        choices = self.coloring.selectCompatibleChoices(family.family)
        mdp,state_map,choice_map = self.restrict_quotient(choices)
        pomdp = self.obs_evaluator.add_observations_to_submdp(mdp,state_map)
        # for state,quotient_state in enumerate(state_map):
        #     assert pomdp.observations[state] == self.state_to_observation[quotient_state]
        # assert pomdp.nr_observations == self.num_observations
        return SubPomdp(pomdp,self,state_map,choice_map)


    def build_dtmc_sketch(self, fsc, negate_specification=True):
        '''
        Construct the family of DTMCs representing the execution of the given FSC in different environments.
        '''

        # create the product
        fsc.check(self.observation_to_actions)
        
        self.fsc_unfolder = payntbind.synthesis.FscUnfolder(
            self.quotient_mdp, self.state_to_observation, self.num_actions, self.choice_to_action
        )
        if isinstance(fsc,paynt.quotient.fsc.Fsc):
            self.fsc_unfolder.applyFsc(fsc.transitions)
        elif isinstance(fsc,paynt.quotient.fsc.FscFactored):
            self.fsc_unfolder.applyFscFactored(fsc.action_function, fsc.update_function)
        else:
            raise ValueError("unknown FSC class")
        product = self.fsc_unfolder.product
        product_choice_to_choice = self.fsc_unfolder.product_choice_to_choice

        # the product inherits the design space
        product_family = self.family.copy()
        
        # the choices of the product inherit colors of the quotient
        product_choice_to_hole_options = []
        quotient_num_choces = self.quotient_mdp.nr_choices
        choice_to_hole_assignment = self.coloring.getChoiceToAssignment()
        for product_choice in range(product.nr_choices):
            choice = product_choice_to_choice[product_choice]
            if choice == quotient_num_choces:
                hole_options = []
            else:
                hole_options = [(hole,option) for hole,option in choice_to_hole_assignment[choice]]
            product_choice_to_hole_options.append(hole_options)
        product_coloring = payntbind.synthesis.Coloring(product_family.family, product.nondeterministic_choice_indices, product_choice_to_hole_options)
        
        # copy the specification
        product_specification = self.specification.copy()
        if negate_specification:
            product_specification = product_specification.negate()

        dtmc_sketch = paynt.quotient.quotient.Quotient(product, product_family, product_coloring, product_specification)
        return dtmc_sketch





### LEGACY CODE, NOT UP-TO-DATE ###

    def compute_qvalues_for_product_submdp(self, product_submdp : paynt.models.models.SubMdp):
        '''
        Given an MDP obtained after applying FSC to a family of POMDPs, compute for each state s, (reachable)
        memory node n, and action a, the Q-value Q((s,n),a).
        :note it is assumed that a randomized FSC was used
        :note it is assumed the provided DTMC sketch is the one obtained after the last unfolding, i.e. no other DTMC
            sketch was constructed afterwards
        :return a dictionary mapping (s,n,a) to Q((s,n),a)
        '''
        assert isinstance(self.product_pomdp_fsc, payntbind.synthesis.ProductPomdpRandomizedFsc), \
            "to compute Q-values, unfolder for randomized FSC must have been used"

        # model check
        prop = self.get_property()
        result = product_submdp.model_check_property(prop)
        product_state_sub_to_value = result.result.get_values()

        # map states of a sub-MDP to the states of the quotient-MDP to the state-memory pairs of the quotient POMDP
        # map states values to the resulting map
        product_state_to_state_memory_action = self.product_pomdp_fsc.product_state_to_state_memory_action.copy()
        state_memory_action_to_value = {}
        invalid_action = self.num_actions
        for product_state_sub in range(product_submdp.model.nr_states):
            product_state = product_submdp.quotient_state_map[product_state_sub]
            state,memory_action = product_state_to_state_memory_action[product_state]
            memory,action = memory_action
            if action == invalid_action:
                continue
            value = product_state_sub_to_value[product_state_sub]
            state_memory_action_to_value[(state,memory,action)] = value
        return state_memory_action_to_value


    def translate_path_to_trace(self, dtmc, path):
        invalid_choice = self.quotient_mdp.nr_choices
        trace = []
        for dtmc_state in path:
            product_choice = dtmc.quotient_choice_map[dtmc_state]
            choice = self.product_pomdp_fsc.product_choice_to_choice[product_choice]
            if choice == invalid_choice:
                # randomized FSC: we are in the intermediate state, move on to the next one
                continue
            
            product_state = dtmc.quotient_state_map[dtmc_state]
            state = self.product_pomdp_fsc.product_state_to_state[product_state]
            obs = self.state_to_observation[state]
            action = self.choice_to_action[choice]
            trace.append( (obs,action) )

        # in the last state, we remove the action since it was not actually used
        trace[-1] = (obs,None)

        # sanity check
        for obs,action in trace[:-1]:
            assert action in self.observation_to_actions[obs], "invalid trace"

        return trace
    
    
    def compute_witnessing_traces(self, dtmc_sketch, satisfying_assignment, num_traces, trace_max_length):
        '''
        Generate witnessing paths in the DTMC induced by the DTMC sketch and a satisfying assignment.
        If the set of target states is not reachable, then random traces are simulated
        :return a list of state-action pairs
        :note the method assumes that the DTMC sketch is the one that was last constructed using build_dtmc_sketch()
        '''
        fsc_is_randomized = isinstance(self.product_pomdp_fsc, payntbind.synthesis.ProductPomdpRandomizedFsc)
        if fsc_is_randomized:
            # double the trace length to account for intermediate states
            trace_max_length *= 2

        # logger.debug("constructing witnesses...")
        dtmc = dtmc_sketch.build_assignment(satisfying_assignment)

        # assuming a single probability reachability property
        spec = dtmc_sketch.specification
        assert spec.num_properties == 1, "expecting a single property"
        prop = spec.all_properties()[0]
        if prop.is_reward:
            logger.warning("WARNING: specification is a reward property, but generated traces \
                will be based on transition probabilities")
        
        target_label = self.extract_target_label()
        target_states = dtmc.model.labeling.get_states(target_label)

        traces = []
        if target_states.number_of_set_bits()==0:
            # target is not reachable: use Stormpy simulator to obtain some random walk in a DTMC
            logger.debug("target is not reachable, generating random traces...")
            simulator = stormpy._DiscreteTimeSparseModelSimulatorDouble(dtmc.model)
            for _ in range(num_traces):
                simulator.reset_to_initial_state()
                path = [simulator.get_current_state()]
                for _ in range(trace_max_length):
                    success = simulator.random_step()
                    if not success:
                        break
                    path.append(simulator.get_current_state())
                trace = self.translate_path_to_trace(dtmc,path)
                traces.append(trace)
        else:
            # target is reachable: use KSP
            logger.debug("target is reachable, computing shortest paths to...")
            if prop.minimizing:
                logger.debug("...target states")
                shortest_paths_generator = stormpy.utility.ShortestPathsGenerator(dtmc.model, target_label)
            else:
                logger.debug("...BSCCs from which target states are unreachable...")
                phi_states = stormpy.storage.BitVector(dtmc.model.nr_states,True)
                states0,_ = stormpy._compute_prob01states_double(dtmc.model,phi_states,target_states)
                shortest_paths_generator = stormpy.utility.ShortestPathsGenerator(dtmc.model, states0)
            for k in range(1,num_traces+1):
                path = shortest_paths_generator.get_path_as_list(k)
                path.reverse()
                trace = self.translate_path_to_trace(dtmc,path)
                traces.append(trace)
        return traces



import sys
import re

def match(regex, lines):
    for line in lines:
        res = re.match(regex, line)
        if res is not None:
            return res.groups()
    return None

# default values
iters_estimate = time_estimate = None
method = synthesis_time = number_of_holes = family_size = None
mdp_size = cegar_iters = None
dtmc_size = cegis_iters = None
ce_quality_maxsat = ce_quality_trivial = ce_quality_nontrivial = None
ce_time_maxsat = ce_time_trivial = ce_time_nontrivial = None
iters = time = None

# process command-line arguments
assert len(sys.argv) == 3
filename = sys.argv[1]
selection = sys.argv[2]

# read file
file = open(filename, 'r')
lines = file.readlines()
file.close()

# we will only need last few lines
lines = lines[-20:]

# timeout check
if lines[-1] == "TO\n":
    res = match(r"^.*?Performance estimation \(unfeasible\): (.*?) iterations in (.*?) sec.*?$", lines)
    if res is not None:
        iters = round(float(res[0]),0)
        time = round(float(res[1]),0)
        iters_estimate = str(iters) + "*"
        time_estimate = str(time) + "*"

res = match(r"^method: (.*?), synthesis time: (.*?) s$", lines)
if res is not None:
    method = res[0]
    synthesis_time = res[1]

res = match(r"^number of holes: (.*?), family size: (.*?)$", lines)
if res is not None:
    number_of_holes = res[0]
    family_size = res[1]

res = match(r"^super MDP size: (.*?), average MDP size: (.*?), MPD checks: (.*?), iterations: (.*?)$", lines)
if res is not None:
    mdp_size = res[1]
    cegar_iters = res[3]

res = match(r"^average DTMC size: (.*?), DTMC checks: (.*?), iterations: (.*?)$", lines)
if res is not None:
    dtmc_size = res[0]
    cegis_iters = res[2]

res = match(r"^ce quality: maxsat: (.*?), trivial: (.*?), non-trivial: (.*?)$", lines)
if res is not None:
    ce_quality_maxsat = res[0]
    ce_quality_trivial = res[1]
    ce_quality_nontrivial = res[2]

res = match(r"^ce time: maxsat: (.*?), trivial: (.*?), non-trivial: (.*?)$", lines)
if res is not None:
    ce_time_maxsat = res[0]
    ce_time_trivial = res[1]
    ce_time_nontrivial = res[2]

hybrid_iters = None if cegar_iters is None or cegis_iters is None else f"({cegar_iters},{cegis_iters})"

# identify iters and time
if method is None:
    # timeout
    iters = iters_estimate
    time = time_estimate
else:
    iters = next(value for value in [hybrid_iters, cegis_iters, cegar_iters] if value is not None)
    time = synthesis_time

# print selected value
value_selected = globals()[selection]
value_str = value_selected if value_selected is not None else "-"
print(value_str)



import paynt.parser.sketch as sketch
import paynt.quotient.mdp
import paynt.quotient.quotient

from helpers.helper import get_stormpy_example_path, get_sketch_paths

class TestSketch:

    def test_load_sketch_mdp_prism(self):
        # setup
        sketch_path, props_path = get_sketch_paths("mdp/maze")

        # test
        quotient = sketch.Sketch.load_sketch(sketch_path, props_path)

        # assert
        assert isinstance(quotient, paynt.quotient.mdp.MdpQuotient)
        assert quotient.quotient_mdp.nr_states == 15
        # TODO add more asserts

    # add more tests